package com.ningdatech.pmapi.todocenter.manage;

import cn.hutool.core.collection.CollUtil;
import com.alibaba.fastjson.JSON;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.ningdatech.basic.exception.BizException;
import com.ningdatech.pmapi.common.statemachine.util.StateMachineUtils;
import com.ningdatech.pmapi.projectdeclared.manage.DeclaredProjectManage;
import com.ningdatech.pmapi.projectdeclared.model.dto.ProjectDraftSaveDTO;
import com.ningdatech.pmapi.projectlib.enumeration.ProjectStatusEnum;
import com.ningdatech.pmapi.projectlib.model.dto.ProjectApplicationDTO;
import com.ningdatech.pmapi.projectlib.model.dto.ProjectDTO;
import com.ningdatech.pmapi.projectlib.model.entity.Project;
import com.ningdatech.pmapi.projectlib.model.entity.ProjectApplication;
import com.ningdatech.pmapi.projectlib.model.entity.ProjectInst;
import com.ningdatech.pmapi.projectlib.service.IProjectApplicationService;
import com.ningdatech.pmapi.projectlib.service.IProjectInstService;
import com.ningdatech.pmapi.projectlib.service.IProjectService;
import com.ningdatech.pmapi.staging.enums.MsgTypeEnum;
import com.ningdatech.pmapi.staging.service.IProjectStagingService;
import com.ningdatech.pmapi.sys.manage.NoticeManage;
import com.ningdatech.pmapi.todocenter.constant.TodoCenterContant;
import com.ningdatech.pmapi.user.service.IUserInfoService;
import com.ningdatech.pmapi.user.util.LoginUserUtil;
import com.wflow.contants.HisProInsEndActId;
import com.wflow.workflow.bean.process.ProgressNode;
import com.wflow.workflow.bean.process.enums.NodeTypeEnum;
import com.wflow.workflow.bean.vo.ProcessProgressVo;
import com.wflow.workflow.enums.ProcessHandlerEnum;
import com.wflow.workflow.service.ProcessInstanceService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.flowable.engine.HistoryService;
import org.flowable.engine.RuntimeService;
import org.flowable.engine.TaskService;
import org.flowable.engine.history.HistoricProcessInstance;
import org.flowable.engine.task.Comment;
import org.flowable.variable.api.history.HistoricVariableInstance;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

import static com.ningdatech.pmapi.todocenter.constant.WorkNoticeContant.*;

/**
 * @Classname HandlerManage
 * @Description
 * @Date 2023/4/2 14:25
 * @Author PoffyZhang
 */
@Component
@RequiredArgsConstructor
@Slf4j
public class HandlerManage {

    private final RuntimeService runtimeService;

    private final TaskService taskService;
    private final HistoryService historyService;
    private final IUserInfoService userInfoService;
    private final IProjectService projectService;
    private final StateMachineUtils stateMachineUtils;
    private final ProcessInstanceService processInstanceService;
    private final IProjectApplicationService projectApplicationService;
    private final IProjectStagingService projectStagingService;
    private final IProjectInstService projectInstService;
    private final NoticeManage noticeManage;
    private final DeclaredProjectManage declaredProjectManage;

    /**
     * 审核通过后 所处理的逻辑
     * @param declaredProject
     * @param instance
     */
    public void afterPassTodo(Project declaredProject, HistoricProcessInstance instance){
        Long userId = LoginUserUtil.getUserId();
        // 获取流程通过后的流程实例
        HistoricProcessInstance newInstance = historyService.createHistoricProcessInstanceQuery()
                .processInstanceId(instance.getId())
                .singleResult();
        // 获取流程通过后当前流程详情
        ProcessProgressVo newInstanceDetail = processInstanceService.getProgressInstanceDetail(null, instance.getId());
        // 获取流程通过后当前审核人信息，向其发送工作通知
        List<ProgressNode> newProgressInfo = newInstanceDetail.getProgressInfo();
        ProgressNode currentNode = newProgressInfo.get(newProgressInfo.size() - 1);
        String currentEmployeeCode;
        // 说明当前节点是子流程节点
        if (currentNode.getNodeType().name().equals(NodeTypeEnum.SUB.name())) {
            List<ProgressNode> children = currentNode.getChildren();
            // 获取子流程当前审核人节点
            ProgressNode subCurrentNode = children.get(children.size() - 1);
            currentEmployeeCode = subCurrentNode.getUserId();
        } else {
            currentEmployeeCode = currentNode.getUserId();
        }

        // 流程通过后，判断当前登录用户是不是最后一个审核人
        // 若当前登录用户是最后一个审批人，需更新流程状态为审核完成，项目状态到下个状态
        // 并向流程发起人发送浙政钉工作通知：【项目名称】已通过【流程名称】，请及时开始下一步操作。
        if (HisProInsEndActId.END.equals(newInstance.getEndActivityId())) {
            switch (Objects.requireNonNull(ProjectStatusEnum.getValue(declaredProject.getStatus()))) {
                // 当前项目状态是预审中
                case PRE_APPLYING:
                    //先修改项目状态
                    updatePassProjectStatus(userId, declaredProject);
                    //然后入库暂存库
                    projectStagingService.addByProject(declaredProject,"暂存入库 待提交部门联审");
                    break;
                // 当前项目状态是单位内部审核中
                case UNDER_INTERNAL_AUDIT:
                    // 当前项目状态是部门联审中
                case DEPARTMENT_JOINT_REVIEW:
                    // 当前项目状态是方案评审中
                case SCHEME_UNDER_REVIEW:
                    // 当前项目状态是终验审核中
                case FINAL_ACCEPTANCE_IS_UNDER_REVIEW:
                    updatePassProjectStatus(userId, declaredProject);
                    break;
                default:
                    throw new IllegalStateException("Unexpected value: " + declaredProject.getStatus());
            }
            //发送消息
            noticeManage.sendNotice(currentEmployeeCode,userId,declaredProject,instance.getProcessDefinitionName(),
                    PASS_MSG_TEMPLATE2,MsgTypeEnum.PROJECT_REVIEW_PASS);
        } else {
            // 若有下一个审核人(当前节点的用户)，
            // 向其发送浙政钉工作通知：标题：审核任务 内容：【单位名称】的【项目名称】需要您审核。
            // 获取发送浙政钉工作通知必要信息
            if (Objects.isNull(currentEmployeeCode)) {
                throw new BizException("审核人信息不存在!");
            }
            //发送消息
            noticeManage.sendNotice(currentEmployeeCode,userId,declaredProject,instance.getProcessDefinitionName(),
                    PASS_MSG_TEMPLATE,MsgTypeEnum.PROJECT_REVIEW);
        }
    }


    /**
     * 驳回后 所处理的逻辑
     * @param declaredProject
     * @param instance
     */
    public void afterRejectTodo(Project declaredProject, HistoricProcessInstance instance) {
        Long userId = LoginUserUtil.getUserId();
        // 更新项目状态和流程状态
        updateRejectProjectStatus(userId, declaredProject);
        //发送消息
        noticeManage.sendNotice(instance.getStartUserId(),userId,declaredProject,instance.getProcessDefinitionName(),
                REJECT_MSG_TEMPLATE,MsgTypeEnum.PROJECT_REVIEW_REJECT);
    }

    /**
     * 发起人 撤回逻辑
     */
    public void rootWithDraw(Project declaredProject) {
        Long userId = LoginUserUtil.getUserId();
        // 若是流程发起人点击撤回，项目回到上一个状态，需调用状态机更新项目状态,流程状态更新为审核通过
        switch (Objects.requireNonNull(ProjectStatusEnum.getValue(declaredProject.getStatus()))) {
            // 当前项目状态是单位内部审核中
            case UNDER_INTERNAL_AUDIT:
                // 更新项目状态为待申报
                updateWithdrawProjectStatus(userId, declaredProject);
                // 保存到草稿箱中
                saveToDraft(declaredProject);
                //删除所有项目相关的信息
                deleteProjectRelated(declaredProject);
                break;
            // 当前项目状态是预审中
            case PRE_APPLYING:
                // 当前项目状态是部门联审中
            case DEPARTMENT_JOINT_REVIEW:
                // 当前项目状态是方案评审中
            case SCHEME_UNDER_REVIEW:
                // 当前项目状态是终验审核中
            case FINAL_ACCEPTANCE_IS_UNDER_REVIEW:
                updateWithdrawProjectStatus(userId, declaredProject);
                break;
            default:
                throw new IllegalStateException("Unexpected value: " + declaredProject.getStatus());
        }
    }

    // 保存到草稿箱中
    private void saveToDraft(Project declaredProject) {
        ProjectDraftSaveDTO draftSaveDto = new ProjectDraftSaveDTO();
        ProjectDTO projectInfo = new ProjectDTO();
        BeanUtils.copyProperties(declaredProject,projectInfo);
        // 查询出项目关联的应用信息
        List<ProjectApplication> applicationList = projectApplicationService.list(Wrappers.lambdaQuery(ProjectApplication.class)
                .eq(ProjectApplication::getProjectId, declaredProject.getId()));
        List<ProjectApplicationDTO> applicationDTOList = applicationList.stream().map(a -> {
            ProjectApplicationDTO applicationDTO = new ProjectApplicationDTO();
            BeanUtils.copyProperties(a, applicationDTO);
            return applicationDTO;
        }).collect(Collectors.toList());
        projectInfo.setApplicationList(applicationDTOList);
        HashMap<String,Object> dynamicMap = JSON.parseObject(declaredProject.getDynamicForm(), HashMap.class);
        projectInfo.setDynamicForm(dynamicMap);
        draftSaveDto.setProjectInfo(projectInfo);
        declaredProjectManage.saveToDraft(draftSaveDto);
    }

    /**
     * 删除项目的所有信息
     * @param declaredProject
     */
    public void deleteProjectRelated(Project declaredProject) {
        //删除项目库中该项目信息
        projectService.removeById(declaredProject);
        //删除app信息
        projectApplicationService.remove(Wrappers.lambdaQuery(ProjectApplication.class)
                .eq(ProjectApplication::getProjectId,declaredProject.getId()));
    }

    /**
     * 退回审核后 所处理的逻辑
     * @param declaredProject
     * @param instance
     */
    public void afterBackTodo(Project declaredProject, HistoricProcessInstance instance) {
        Long userId = LoginUserUtil.getUserId();
        // 给项目创建人、流程发起人发送浙政钉工作通知：【项目名称】的【流程名称】被退回，请及时处理。
        //发送消息
        noticeManage.sendNotice(instance.getStartUserId(),userId,declaredProject,instance.getProcessDefinitionName(),
                BACK_MSG_TEMPLATE,MsgTypeEnum.PROJECT_REVIEW_BACK);
    }


    /**
     * 当为通过操作时，更新项目表中项目状态
     *
     * @param userId
     * @param declaredProject
     * @return void
     * @author CMM
     * @since 2023/02/08
     */
    private void updatePassProjectStatus(Long userId, Project declaredProject) {
        try {
            stateMachineUtils.pass(declaredProject);
            declaredProject.setUpdateOn(LocalDateTime.now());
            declaredProject.setUpdateBy(userId);
            projectService.updateById(declaredProject);
        } catch (Exception e) {
            log.error("状态机执行失败",e);
            throw new BizException("状态机执行失败!");
        }
    }

    /**
     * 当为驳回操作时，更新项目表中的项目状态
     *
     * @param userId
     * @param declaredProject
     * @return void
     * @author CMM
     * @since 2023/02/08
     */
    private void updateRejectProjectStatus(Long userId, Project declaredProject) {
        stateMachineUtils.reject(declaredProject);
        declaredProject.setUpdateOn(LocalDateTime.now());
        declaredProject.setUpdateBy(userId);
        projectService.updateById(declaredProject);
    }

    /**
     * 当为撤回操作时，更新项目表中的项目状态为前一个状态
     *
     * @param userId
     * @param declaredProject
     * @return void
     * @author CMM
     * @since 2023/02/08
     */
    @Transactional(rollbackFor = Exception.class)
    public void updateWithdrawProjectStatus(Long userId, Project declaredProject) {
        // 删除关联表信息
        projectInstService.remove(Wrappers.lambdaQuery(ProjectInst.class)
                .eq(ProjectInst::getProjectId,declaredProject.getId())
                .eq(ProjectInst::getInstCode,declaredProject.getInstCode()));
        //删除项目的实例信息
        historyService.deleteHistoricProcessInstance(declaredProject.getInstCode());

        stateMachineUtils.withDraw(declaredProject);
        declaredProject.setUpdateOn(LocalDateTime.now());
        declaredProject.setUpdateBy(userId);
        declaredProject.setInstCode(TodoCenterContant.Declared.NULL_INST_CODE);
        projectService.updateById(declaredProject);
    }

    public void deleteBackComments(List<HistoricVariableInstance> approves) {
        if(CollUtil.isNotEmpty(approves)){
            for(HistoricVariableInstance approve : approves){
                if(approve.getValue() instanceof ProcessHandlerEnum){
                    if(ProcessHandlerEnum.BACK.equals(ProcessHandlerEnum.getEnumByName(String.valueOf(approve.getValue())))){
                        runtimeService.removeVariable(approve.getProcessInstanceId(),approve.getVariableName());

                        Comment comment;
                        Optional<Comment> first = taskService.getProcessInstanceComments(approve.getProcessInstanceId())
                                .stream()
                                .filter(c -> c.getTaskId().equals(approve.getVariableName().replace("approve_", StringUtils.EMPTY)))
                                .findFirst();
                        if (first.isPresent()){
                            comment = first.get();
                            taskService.deleteComment(comment.getId());
                        }

                        break;
                    }
                }
            }
        }
    }
}
